--
-- TR_Gameplay.lua
-- Functions for how this mode is supposed to... well... function.
-- 
-- By Flame
-- 5/10/20
--

TR.Game = {}

-- Detect what control scheme we're using
TR.Game.ControlStyle = function(p)
	-- Returns 6 if "Simple" control scheme is enabled.
	-- Returns 4 if directionchar is set to movement
	-- Returns 2 if directionchar is set to camera, but Simple control scheme is enabled, AKA Analog Legacy
	-- Returns 0 if directionchar is set to camera, AKA Legacy
	return ((p.pflags & PF_ANALOGMODE) and 2 or 0) | ((p.pflags & PF_DIRECTIONCHAR) and 4 or 0)
end

TR.Game.drainAmmo = function(p, power, amount)
	if amount
		p.powers[power] = $ - amount
		p.rings = $ - amount
		
		if (p.powers[power] < 0)
			p.powers[power] = 0
		end
		
		if p.rings < 0
			p.rings = 0
		end
	else
		p.powers[power] = $ - 1
		p.rings = $ - 1
	end
	-- Copied from SRB2 Source
	/*if (p.rings < 1)
		p.ammoremovalweapon = p.currentweapon
		p.ammoremovaltimer = ammoremovaltics -- 2*TICRATE
		
		if (p.powers[power] > 0)
			p.powers[power] = $ - 1
			p.ammoremoval = 2
		else
			p.ammoremoval = 1
		end
	else
		p.rings = $ - 1
	end*/
end

TR.Game.spawnFastTrail = function(thing)
	if thing and thing.valid
		local r = (thing.radius*2)>>FRACBITS
		local trail = P_SpawnMobj(
				thing.x + (P_RandomRange(r, -r)<<FRACBITS),
				thing.y + (P_RandomRange(r, -r)<<FRACBITS),
				thing.z + (P_RandomKey(thing.height>>FRACBITS)<<FRACBITS),
				MT_FASTTRAIL)
		trail.target = thing
		trail.angle = thing.angle
		trail.momx = thing.momx/2
		trail.momy = thing.momy/2
		trail.scalespeed = FRACUNIT/20
		trail.destscale = 1
		--trail.angle = $ - ANGLE_180
	end
end

TR.Game.spawnRespawnObj = function(thing)
	if thing and thing.valid
		S_StartSound(thing, sfx_s3kcas)
		for i = 1, 8 do
			local fa = (i*ANGLE_45)
			local dist = thing.radius*2+(FRACUNIT*2)
			local r = P_SpawnMobj(thing.x + FixedMul(cos(fa), dist),
								thing.y + FixedMul(sin(fa), dist), 
								thing.z, MT_RESPAWNFX)
			r.angle = fa+ANGLE_90
			r.fuse = 8
			P_SetObjectMomZ(r, FRACUNIT*8, false)
		end
	end
end

TR.Game.spawnWidePhantom = function(thing)
	if thing and thing.valid
		local r = thing.radius*4
		for i = -1, 1, 1
			if i != 0 -- When i is a non-zero value... Do this!
				local phantom = P_SpawnMobj(thing.x 
											+ P_ReturnThrustX(thing, thing.angle, r)
											+ P_ReturnThrustY(thing, thing.angle, i*r)
											- FixedMul(cos(thing.angle), r*2),
											
											thing.y
											+ P_ReturnThrustY(thing, thing.angle, r) 
											+ P_ReturnThrustX(thing, thing.angle, i*r)
											- FixedMul(sin(thing.angle), r*2),
											
											thing.z,
											MT_PHANTOMTRAIL)

				phantom.target = thing
				phantom.angle = thing.angle
				phantom.skin = thing.skin
				phantom.sprite2 = thing.sprite2
				phantom.frame = thing.frame | (TR_TRANS50|FF_FULLBRIGHT) & ~(FF_ANIMATE)
				
				phantom.color = thing.color
				phantom.fuse = 2
			end
		end
	end
end

TR.Game.spawnOrbitars = function(thing)
	if thing and thing.valid
		local r = thing.radius*4
		local orbitar1 = P_SpawnMobj(thing.x + thing.momx + FixedMul(cos(thing.angle), r*2),
									thing.y + thing.momy + FixedMul(sin(thing.angle), r*2),
									thing.z,
									MT_TRONTRAIL)
		orbitar1.target = thing
		orbitar1.momx = thing.momx
		orbitar1.momy = thing.momy
		orbitar1.angle = thing.angle
		orbitar1.color = thing.color
		orbitar1.threshold = 2
		orbitar1.reactiontime = 1 -- This is purely for sound related purposes
		orbitar1.fuse = TICRATE*5
		
		local orbitar2 = P_SpawnMobj(thing.x - FixedMul(cos(thing.angle), r*2),
									thing.y - FixedMul(sin(thing.angle), r*2),
									thing.z,
									MT_TRONTRAIL)
		orbitar2.target = thing
		orbitar2.momx = thing.momx
		orbitar2.momy = thing.momy
		orbitar2.angle = thing.angle+ANGLE_180
		orbitar2.color = thing.color
		orbitar2.threshold = 2
		orbitar2.reactiontime = 2 -- This is purely for sound related purposes
		orbitar2.fuse = TICRATE*5
	end
end

TR.Game.spawnBlastShield = function(thing)
	if thing and thing.valid
		local shield = P_SpawnMobj(thing.x + thing.momx,
									thing.y + thing.momy, 
									thing.z, 
									MT_KAMIKAZESHIELD)
		shield.target = thing
		shield.angle = thing.angle
		shield.color = thing.color
		shield.fuse = -1
	end
end

TR.Game.spawnTronTrail = function(thing)
	if thing and thing.valid
		local zpos = P_FloorzAtPos(thing.x,
									thing.y,
									thing.z,
									thing.height)
									
		local trail = P_SpawnMobj(thing.x,
									thing.y,
									zpos, 
									MT_TRONTRAIL)
		
		-- Award score properly
		if thing.player -- Did a real player spawn this?
			trail.target = thing -- This is your target.
		else
			trail.target = thing.target -- Otherwise, your mo's target is also your target.
		end
		trail.angle = R_PointToAngle2(trail.x, trail.y, thing.x + thing.momx, thing.y + thing.momy) -- Always point to your target
		trail.scale = 1 -- Start with this scale
		trail.destscale = (2*FRACUNIT)-(FRACUNIT/2) -- Grow to this scale (1.5)
		
		-- Grow to this scale in this amount of time
		trail.scalespeed = FRACUNIT/24 -- Normal scale time (Slower)
		
		-- Kill the Mobj after this amount of time
		if thing.player
			trail.fuse = TICRATE* (5 - thing.player.killstreak or 0)
		else
			trail.fuse = TICRATE*4 
		end
		trail.color = trail.target.color -- Color
		--trail.flags2 = MF2_DONTDRAW -- Don't draw this.
	end
end

TR.Game.dotronTrail = function(p)
	local mo = p.mo
	local zpos = P_FloorzAtPos(mo.x,
								mo.y,
								mo.z,
								mo.height)
								
	-- Player is on ground?
	-- Player is going faster than a speed of 8?
	-- Spawn a trail!
	if (p.speed >= 8*FRACUNIT)
		if P_IsObjectOnGround(mo)
			TR.Game.spawnTronTrail(mo)
		elseif (leveltime%4 == 0) -- If you're in the air, spawn clouds!
			local trail = P_SpawnMobj(mo.x,
										mo.y,
										mo.z + P_GetPlayerHeight(p)/2, 
										MT_TNTDUST)
			trail.angle = R_PointToAngle2(trail.x, trail.y, mo.x + mo.momx, mo.y + mo.momy) -- Always point to your target
			trail.destscale = 2*FRACUNIT
			trail.scalespeed = FRACUNIT/(3*TICRATE/2)
			trail.flags = $ | MF_NOGRAVITY
		end
	end
end

TR.Game.tronPreThink = function(p)
	if not p.powers[pw_nocontrol] -- If you have control.
		p.cmd.forwardmove = 50 -- Always move forward
	end
	if (p.weapondelay <= 2)
		p.weapondelay = 2 -- Do not fire rings. Ever.
	end
	if (p.cmd.buttons & BT_USE)
		p.cmd.buttons = $ | (BT_ATTACK) & ~(BT_USE)
	end
	p.cmd.angleturn = p.mo.angle>>16 -- Set your turning to wherever you're facing
end

TR.Game.tronRespawn = function(p)
	-- Give player the following:
	-- Set player.powers[pw_nocontrol] timer to the respawn flashing timer.
	-- Set player 4x above spawn point
	-- player.justrespawned flag, No gravity
	local spawndelay = 3*TICRATE
	p.respawntimer = spawndelay
	p.powers[pw_nocontrol] = spawndelay
	p.mo.flags = $ | MF_NOGRAVITY
	p.chargetimer = 0
	p.killstreak = 0
	p.kamialert = false
	P_TeleportMove(p.mo, p.mo.x, p.mo.y, p.mo.z + P_GetPlayerHeight(p)*4)
end

TR.Game.tronThink = function(p)
	local cmd = p.cmd
	local mo = p.mo
	local speed = p.normalspeed
	
	-- Remove Dashmode stuff. Way too dumb in this mode!
	if (p.charflags & SF_DASHMODE)
		p.charflags = $ & ~(SF_DASHMODE)
	end
	
	-- Respawn timer stuff
	if (p.respawntimer > 0) -- Respawn timer active!
		p.respawntimer = $ - 1 -- Decrease by one!
		if (p.respawntimer <= (3*TICRATE)/4) -- At 3/4th's of a tic,
			if (p.respawntimer == (3*TICRATE)/4) -- Trigger this only once:
				if (mo.flags & MF_NOGRAVITY) -- Regain your gravity
					mo.flags = $ & ~MF_NOGRAVITY
				end
				-- And get ready to zoom!
				p.panim = PA_RUN
				mo.state = S_PLAY_DASH
				S_StartSound(mo, sfx_spin) -- Rev
			elseif (p.respawntimer == 1) -- Last tic!
				S_StopSound(mo)
				S_StartSound(mo, sfx_s3kb6) -- Zoom
				--S_StartSound(mo, sfx_kc3c) -- Zoom
			end
		else -- Otherwise you're 'standing'.
			p.panim = PA_IDLE
			mo.state = S_PLAY_STND
			p.powers[pw_flashing] = (3*TICRATE/2)
		end
	end
	
	 -- Respawning?
	if p.powers[pw_flashing] and (leveltime%8 == 0) and (p.respawntimer > (3*TICRATE)/4) then TR.Game.spawnRespawnObj(mo) end
	
	-- Kill streak stuff
	if (p.killstreak > 3)
		p.killstreak = 3 -- Max of 3!
	end
	
	-- Dashing speed?
	if p.powers[pw_sneakers]
		speed = ($ * 5)/ 3 -- 1.67x
		TR.Game.spawnFastTrail(mo)
	end
	
	if (p.rings > 0) -- Player has rings?
		if (cmd.buttons & BT_ATTACK) and not (p.pflags & PF_ATTACKDOWN) -- Pressing the attack button
			p.pflags = $ | PF_ATTACKDOWN
			if (p.currentweapon == WEP_AUTO) -- Automatic Ring (Speed Burst) selected
			and (p.ringweapons & RW_AUTO) -- Weapon ring able to be fired
			and (p.powers[pw_automaticring] > 0) -- Player has Auto rings to spare?
			and (p.weapondelay <= 4)
				-- Let's give your character a dashing ability
				p.pflags = $ & ~PF_ATTACKDOWN -- Attack is repeatable
				if (leveltime%(TICRATE/2) == 0) -- Every half a second...
					TR.Game.drainAmmo(p, pw_automaticring) -- Decrease your weapon ring count by one
					S_StartSound(mo, sfx_antiri) -- Play a sound...
				end

				if not p.powers[pw_flashing] and not p.powers[pw_sneakers] -- Just starting from a dash...
					p.powers[pw_flashing] = TICRATE/4
					S_StopSound(mo)
					--S_StartSound(mo, sfx_mswarp) -- Zoom!
					S_StartSound(mo, sfx_s3kb6) -- Spin Launch
				end
				p.powers[pw_sneakers] = 2
				
			elseif (p.currentweapon == WEP_SCATTER) -- Scatter Ring (Phantom Trail) is selected
			and (p.ringweapons & RW_SCATTER) -- Weapon ring able to be fired
			and (p.powers[pw_scatterring] > 0) -- Player has Scatter rings to spare?
			and (p.weapondelay <= 4)
				-- Let's give your character a wide phantom trail mobj!
				p.pflags = $ & ~PF_ATTACKDOWN -- Attack is repeatable
				if (leveltime%(TICRATE/2) == 0) -- Every half a second...
					TR.Game.drainAmmo(p, pw_scatterring) -- Decrease your weapon ring count by one
					S_StartSound(mo, sfx_antiri) -- Play a sound...
				end
				
				if not p.powers[pw_flashing] and (p.weapondelay <= 2)
					S_StopSound(mo)
					S_StartSound(mo, sfx_3db16) -- Play a sound!
				end
				TR.Game.spawnWidePhantom(mo)
				p.weapondelay = 4
				
			elseif (p.currentweapon == WEP_GRENADE) -- Grenade Ring (Orbitars) is selected
			and (p.ringweapons & RW_GRENADE)
			and (p.powers[pw_grenadering] > 0)
			and p.weapondelay <= 2
				TR.Game.drainAmmo(p, pw_grenadering)
				if not p.powers[pw_flashing] and (p.weapondelay <= 2)
					S_StopSound(mo)
					S_StartSound(mo, sfx_3db16) -- Play a sound!
				end
				TR.Game.spawnOrbitars(mo)
				p.weapondelay = 5*TICRATE
				
			elseif (p.currentweapon == WEP_EXPLODE) -- Explosion Ring (Kamikaze) is selected
			and (p.ringweapons & RW_EXPLODE)
			and (p.powers[pw_explosionring] > 0)
			and p.weapondelay <= 2
				if not p.powers[pw_flashing] and (p.weapondelay <= 2)
					S_StopSound(mo)
					S_StartSound(mo, sfx_s3k53) -- Charge up your explosion
					TR.Game.spawnBlastShield(mo)
				end
				p.weapondelay = 5*TICRATE
			end
		end
	end
	
	/*if (cmd.buttons & BT_CUSTOM2) -- Debug
		p.rings = 90
		p.powers[pw_automaticring] = 12
		p.powers[pw_scatterring] = 10
		p.powers[pw_grenadering] = 3
		p.powers[pw_bouncering] = 5
		p.powers[pw_explosionring] = 1
		p.ringweapons = $ | RW_AUTO|RW_SCATTER|RW_GRENADE|RW_EXPLODE
	end*/
	
	/*if (p.rings > 0) -- Player has rings?
		if (leveltime%(TICRATE/2) == 0) -- Every second...
			p.rings = $ - 1 -- Decrease by one.
			S_StartSound(mo, sfx_antiri) -- Play a sound...
			p.powers[pw_flashing] = $ + (TICRATE/2) -- But you gain invulnerability!
		end
	end*/
	
	if not p.cmd.sidemove
		p.directionmove = false
	elseif not p.directionmove
		if (p.cmd.sidemove > 0)
			mo.angle = $ - ANGLE_90 -- Turn right!
			p.drawangle = mo.angle
			p.directionmove = true
		elseif (p.cmd.sidemove < 0)
			mo.angle = $ + ANGLE_90 -- Turn left!
			p.drawangle = mo.angle
			p.directionmove = true
		end
	end
	
	if not p.powers[pw_nocontrol] -- If you have control
		P_InstaThrust(mo, mo.angle, speed) -- Move Forward
	end
end

TR.Game.doLineCollide = function(p)
	if p and p.valid
	and p.mo and p.mo.valid -- Valid check
		local mo = p.mo -- Simplify ourselves
		
		if (p.rings > 0) -- Ring check.
		and (p.powers[pw_bouncering] > 0) -- Do you have at least one bounce ring in your arsenal?
			-- Then proceed with bouncing off the wall!
			S_StartSound(mo, sfx_s3k61) -- 'Drilling' sfx?
			S_StartSound(mo, sfx_antiri)  -- Play a sound.
			P_TeleportMove(mo, mo.x, mo.y, mo.z + P_GetPlayerHeight(p)+FRACUNIT)
			mo.angle = $ + ANGLE_180
			P_SetObjectMomZ(mo, 5*FRACUNIT, false)
			if (p.cmd.forwardmove != 50) -- If by some miracle you bump into a wall and you're not holding forward...
				P_InstaThrust(mo, mo.angle, p.normalspeed) -- Thurst yourself forward.
			end
			mo.state = S_PLAY_ROLL
			mo.panim = PA_ROLL -- Animate properly

			p.powers[pw_flashing] = TICRATE/2
			TR.Game.drainAmmo(p, pw_bouncering)
			--S_StartSound(mo, sfx_s3kaa) -- 'Bumper'
		else
			p.killstreak = 0
			-- You just crashed!
			P_DamageMobj(mo, nil, nil, 1, DMG_DEATHMASK) -- Die
			mo.momx = 0
			mo.momy = 0
			mo.momz = 0
			mo.flags = $ | (MF_NOGRAVITY) & ~(MF_NOCLIPHEIGHT)
			mo.flags2 = $ | MF2_DONTDRAW
		end
	end
end

TR.Game.tronLineCollide = function(mo, line)
	if (gametype == GT_TRON)
		if mo and mo.valid and line and line.valid -- Mo check, Line check
		and mo.player and mo.player.valid and not mo.player.spectator -- Player check...
			local p = mo.player -- Simplify ourselves
			
			if (
				(line.frontsector and line.frontsector.valid
					and not line.frontsector.f_slope
					and (line.frontsector.floorheight > (mo.z + MAXSTEPMOVE)))
				or (line.backsector and line.backsector.valid 
					and not line.backsector.f_slope
					and (line.backsector.floorheight > (mo.z + MAXSTEPMOVE)))
				) -- Too great to step-up?

				if (p.speed >= p.runspeed) -- Running too fast?
				and (p.playerstate != PST_DEAD) -- Not already dead?
				and (mo.state != S_PLAY_DEAD)
					TR.Game.doLineCollide(p)
				end
			else -- Collision with something else?
				-- Let's check for an FOF.
				local px, py, pz, checksec
				px = mo.x + p.rmomx + FixedMul(cos(mo.angle), mo.radius)
				py = mo.y + p.rmomy + FixedMul(sin(mo.angle), mo.radius)
				if mo.eflags & MFE_VERTICALFLIP
					pz = mo.z + mo.height - FRACUNIT
				else
					pz = mo.z + FRACUNIT
				end
				checksec = R_PointInSubsector(px, py).sector
				
				for rover in checksec.ffloors()
					if not (rover.flags & FF_EXISTS) -- Doesn't exist?
						continue -- Skip it
					end
					
					if not (rover.flags & FF_SOLID) -- Not solid?
						continue -- Skip it
					end
					
					-- Z Height checks
					if not ((pz < rover.topheight) and (pz > rover.bottomheight)) -- Not in range?
						continue -- Skip it
					end
					
					-- Repeat of the above code
					if (p.speed >= p.runspeed) -- Running too fast?
					and (p.playerstate != PST_DEAD) -- Not already dead?
					and (mo.state != S_PLAY_DEAD) 
					and rover.topheight > (mo.z + MAXSTEPMOVE) -- Too great to step-up?
					and not rover.t_slope
						TR.Game.doLineCollide(p)
					end
				end
			end
		end
	end
end

TR.Game.tronSpecialTouch = function(special, toucher)
	if special and special.valid
	and toucher and toucher.valid
	and toucher.player and toucher.player.valid
		local p = toucher.player
	
		if (special.type == MT_BOUNCERING)
		or (special.type == MT_RAILRING)
		or (special.type == MT_INFINITYRING)
		or (special.type == MT_AUTOMATICRING)
		or (special.type == MT_EXPLOSIONRING)
		or (special.type == MT_SCATTERRING)
		or (special.type == MT_GRENADERING)
		or (special.type == MT_BOUNCEPICKUP)
		or (special.type == MT_RAILPICKUP)
		or (special.type == MT_AUTOPICKUP)
		or (special.type == MT_EXPLODEPICKUP)
		or (special.type == MT_SCATTERPICKUP)
		or (special.type == MT_GRENADEPICKUP)
			if not P_CanPickupItem(p, true)
				return false -- Just do default behavior
			end
			
			if special.info.reactiontime -- Panel Pickups
				--p.powers[special.info.mass] = $ + special.reactiontime
				
				if (special.type == MT_AUTOPICKUP) -- Auto -> Super Speed
					if not (p.ringweapons & RW_AUTO) -- Not set.
						p.ringweapons = $ | RW_AUTO -- Set it.
					end
					p.powers[special.info.mass] = $ + 3
					if (p.powers[special.info.mass] > 12)
						p.powers[special.info.mass] = 12 -- Max of 12
					end
				elseif (special.type == MT_BOUNCEPICKUP) -- Bounce -> Wall Safety (Passive)
					-- NEVER Make this Bounce ability availabile. It's PASSIVE for a reason.
					--p.ringweapons = $ | RW_BOUNCE
					p.powers[special.info.mass] = $ + 1
					if (p.powers[special.info.mass] > 3)
						p.powers[special.info.mass] = 3 -- Max of 3
					end
				elseif (special.type == MT_RAILPICKUP) -- Rail - Not available yet.
					p.ringweapons = $ & ~RW_RAIL -- I SAID NOT YET
					p.powers[special.info.mass] = 0
				elseif (special.type == MT_EXPLODEPICKUP) -- Explosion -> Kamikaze
					if not (p.ringweapons & RW_EXPLODE) -- Not set?
						p.ringweapons = $ | RW_EXPLODE -- Set it.
					end
					p.powers[special.info.mass] = $ + 1
					if (p.powers[special.info.mass] > 1)
						p.powers[special.info.mass] = 1 -- Max of 1 (One charge)
					end
				elseif (special.type == MT_SCATTERPICKUP) -- Scatter -> Phantom Trail
					if not (p.ringweapons & RW_SCATTER) -- Not set?
						p.ringweapons = $ | RW_SCATTER -- Set it.
					end
					p.powers[special.info.mass] = $ + 4
					if (p.powers[special.info.mass] > 10)
						p.powers[special.info.mass] = 10 -- Max of 10
					end
				elseif (special.type == MT_GRENADEPICKUP) -- Grenade -> Orbitars
					if not (p.ringweapons & RW_GRENADE) -- Not set?
						p.ringweapons = $ | RW_GRENADE -- Set it
					end
					p.powers[special.info.mass] = $ + 1
					if (p.powers[special.info.mass] > 3)
						p.powers[special.info.mass] = 3 -- Max of 3
					end
				end
			elseif special.info.spawnhealth -- Ammo Pickups.
				--p.powers[special.info.mass] = $ + special.health
				
				if (special.type == MT_BOUNCERING) -- Bounce -> Wall Safety (Passive)
					-- NEVER Make this Bounce ability availabile. It's PASSIVE for a reason.
					p.powers[special.info.mass] = $ + 1
					if (p.powers[special.info.mass] > 3)
						p.powers[special.info.mass] = 3 -- Max of 3
					end
				elseif (special.type == MT_RAILRING) -- Rail - Not available yet.
					p.powers[special.info.mass] = 0 -- I SAID NOT YET
				--elseif (special.type == MT_INFINITYRING)
				elseif (special.type == MT_AUTOMATICRING) -- Auto -> Super Speed
					p.powers[special.info.mass] = $ + 6
					if (p.powers[special.info.mass] > 12)
						p.powers[special.info.mass] = 12 -- Max of 12
					end
				elseif (special.type == MT_EXPLOSIONRING) -- Explosion -> Kamikaze
					p.powers[special.info.mass] = $ + 1
					if (p.powers[special.info.mass] > 1)
						p.powers[special.info.mass] = 1 -- Max of 1 (One charge)
					end
				elseif (special.type == MT_SCATTERRING) -- Scatter -> Phantom Trail
					p.powers[special.info.mass] = $ + 5
					if (p.powers[special.info.mass] > 10)
						p.powers[special.info.mass] = 10 -- Max of 10
					end
				elseif (special.type == MT_GRENADERING) -- Grenade -> Orbitars
					p.powers[special.info.mass] = $ + 3
					if (p.powers[special.info.mass] > 3)
						p.powers[special.info.mass] = 3 -- Max of 3
					end
				end
				p.rings = $ + 1 -- Grant a ring, why not?
			end
		end
	end
	
	S_StartSound(toucher, special.info.deathsound)
	P_KillMobj(special, nil, toucher)
	return true -- Override default behavior.
end

TR.Game.tronThingCollide = function(thing, tmthing)
	-- tmthing is our colliding object
	-- thing is our object that is colliding into tmthing
	if thing and thing.valid
	and tmthing and tmthing.valid -- Valid checks
	and thing.player and thing.player.valid -- Valid thing.player check.
	and not thing.player.spectator
	and (thing.player.playerstate != PST_DEAD) -- thing.player not already dead?
	and (thing.state != S_PLAY_DEAD)
	and not thing.player.powers[pw_flashing] -- Be a little generous with spawning.
		local p = thing.player
		
		if (thing.z >= (tmthing.z + tmthing.height)) -- No Z collision? Let's fix that!
		or ((thing.z + P_GetPlayerHeight(thing.player)) < tmthing.z)
			return -- Out of range
		end
		
		-- Let's get to the OBJECT
		if (tmthing.type == MT_TRONTRAIL)
		and not (tmthing.threshold) -- Not a "Smoking" object
		or ((tmthing.threshold) and tmthing.target and (tmthing.target != thing)) -- Threshold exception
			if (p.powers[pw_bouncering] > 0)
				--S_StartSound(mo, sfx_s3ka8) -- 'Boing' sfx
				S_StartSound(mo, sfx_s3k92) -- 'Ghost' sfx
				S_StartSound(mo, sfx_antiri) -- Play a sound
				TR.Game.drainAmmo(p, pw_bouncering)
				p.powers[pw_flashing] = TICRATE/2
			else
				if tmthing.target and (tmthing.target != thing)
					P_DamageMobj(thing, tmthing, tmthing.target, 1, DMG_DEATHMASK)
					if tmthing.target.player
						tmthing.target.player.killstreak = $ + 1
					end
				else
					P_DamageMobj(thing, nil, nil, 1, DMG_DEATHMASK)
				end
				thing.momz = 0
				thing.flags = $ | (MF_NOGRAVITY) & ~(MF_NOCLIPHEIGHT)
				thing.flags2 = $ | MF2_DONTDRAW
			end
		elseif (tmthing.flags & MF_SOLID) -- Solid object?
			-- Treat it as if you collided with a line
			TR.Game.doLineCollide(thing.player)
		end
	end
end

TR.Game.KamikazeExplode = function(p, range)
	if p and p.valid
	and p.mo and p.mo.valid
		local mo = p.mo
		-- Spawn Sparks
		for i = 1, 8
			local fa = i*ANGLE_45
			local spark = P_SpawnMobj(mo.x, mo.y, mo.z + mo.height/2, MT_SUPERSPARK)
			spark.angle = fa
			spark.fuse = 10 -- 5 frames, 2 tics each
			P_InstaThrust(spark, spark.angle, 36*FRACUNIT)
			P_FlashPal(p, 2, TICRATE/4)
		end
		
		-- Then proceed with kill. Similar to an Armegeddon explosion
		for player in players.iterate
			if (player == p) continue end -- Skip us
			if (player.playerstate != PST_LIVE) continue end -- Skip anyone not alive
			if not player.mo continue end -- Not a valid mo object
			
			-- Calculate distance
			local dist = abs(P_AproxDistance(P_AproxDistance(mo.x - player.mo.x, mo.y - player.mo.y), mo.z - player.mo.z))
			
			if (dist < range) -- In range!
				--print("Player "+player.name+" is in range!")
				--print("Player range is: "+dist/FRACUNIT)
				P_FlashPal(player, 2, TICRATE/4)
				P_DamageMobj(player.mo, mo, mo, 1, DMG_DEATHMASK)
				p.killstreak = $ + 1
			end
		end
	end
end

TR.Game.WarnNearbyPlayersInRange = function(p, range)
	if p and p.valid
	and p.mo and p.mo.valid
		local mo = p.mo
		-- For players in range of the Kamikaze explosion, alert them in some fashion!
		for player in players.iterate
			if (player == p) continue end -- Skip us
			if (player.playerstate != PST_LIVE) continue end -- Skip anyone not alive
			if player.kamialert continue end -- Skip already alerted players
			if not player.mo continue end -- Not a valid mo object
			
			-- Calculate distance
			local dist = abs(P_AproxDistance(P_AproxDistance(mo.x - player.mo.x, mo.y - player.mo.y), mo.z - player.mo.z))
			
			if (dist < range) -- In range!
				player.kamialert = true
				--print("Player "+player.name+" is in range!")
				local alert = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_KAMIKAZEALERT)
				alert.target = player.mo
				alert.fuse = TICRATE/2
				alert.angle = player.mo.angle
				alert.scale = 3*FRACUNIT/2
			end
		end
	end
end

TR.Game.tronDeath = function(mo)
	if mo and mo.valid
	and mo.player and mo.player.valid
		local p = mo.player
		/*for i = 1, 16 do
			local fa = i*ANGLE_22h
			local fire = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SPINFIRE)
			fire.angle = fa
			fire.fuse = TICRATE*2
			--fire.flags = $ & ~MF_NOGRAVITY
			P_SetObjectMomZ(fire, FRACUNIT*2, false)
			P_Thrust(fire, fire.angle, FRACUNIT*3)
		end*/
		p.chargetimer = 0
		p.killstreak = 0
		p.kamialert = false

		S_StopSound(mo)
		if mo.player --and P_IsObjectOnGround(mo)
			P_ElementalFire(p, true)
		end
		
		-- Spawn the "Smoke spawner"
		local sspawn = P_SpawnMobj(mo.x, mo.y, mo.z, MT_TRONTRAIL) -- We can repurpose the Tron trail for something else! :)
		--sspawn.target = mo
		sspawn.threshold = 1 -- And here's how we repurpose it!!
		sspawn.fuse = 7*TICRATE
		sspawn.flags2 = $ | MF2_DONTDRAW
		
		-- Poof!
		local poof = P_SpawnMobj(mo.x, mo.y, mo.z, MT_EXPLODE)
		P_SetMobjStateNF(poof, S_FBOMB_EXPL1)
		S_StartSound(poof, sfx_s3k4e)
		return true
	end
end

-- Our custom thinker objects.
TR.Game.tronThingThinker = function(mo)
	if mo and mo.valid
		if (mo.type == MT_TRONTRAIL)
			if (mo.threshold == 1) and (leveltime%10 == 0) -- Invisible Dust cloud spawner
				local s = P_SpawnMobj(mo.x, mo.y, mo.z, MT_TNTDUST)
				s.angle = P_RandomKey(36)*ANG10
				s.destscale = 2*FRACUNIT
				s.scalespeed = FRACUNIT/TICRATE
				P_SetObjectMomZ(s, FRACUNIT*3, false)
				P_Thrust(s, s.angle, FRACUNIT)
			elseif (mo.threshold == 2) -- pw_grenade ability, orbitar
				if mo.target and mo.target.valid
					if mo.target.player
					and (mo.target.player.playerstate != PST_LIVE) -- Player not alive anymore?
						P_RemoveMobj(mo)
						return
					end
					
					local r = mo.target.radius*8
					mo.angle = $ + ANG10
					P_TeleportMove(mo, mo.target.x + FixedMul(cos(mo.angle), r),
										mo.target.y + FixedMul(sin(mo.angle), r),
										mo.target.z)
					if ((leveltime%TICRATE/4) == 0) and (mo.reactiontime != 2)
						--S_StartSound(mo, sfx_s3k72) -- 
						S_StartSoundAtVolume(mo, sfx_s3kc6s, 150) -- 'Orbiting (S)'
					end
					mo.momx = mo.target.momx
					mo.momy = mo.target.momy
					
					local trail = P_SpawnMobj(mo.x, mo.y, mo.z, MT_TRONTRAIL)
					trail.target = mo.target
					trail.color = mo.color
					trail.momx = mo.momx
					trail.momy = mo.momy
					trail.scale = FRACUNIT
					trail.destscale = 1
					trail.scalespeed = (FRACUNIT/8)
					trail.fuse = 8
					trail.threshold = 3 -- Denotes a trailing object.
					--trail.flags = $ | MF_NOCLIP
				else -- Target disappeared unnaturally?
					P_RemoveMobj(mo) -- You should be gone as well!
				end
			elseif (mo.threshold == 3) -- Orbitar trails
				if mo.target and mo.target.valid
					if (mo.momx != mo.target.momx) or (mo.momy != mo.target.momy) -- Check to see if the player turned 90 degrees
						P_RemoveMobj(mo) -- Remove it if so!
						return
					end
				end
			end
		elseif (mo.type == MT_PHANTOMTRAIL) -- pw_scatter ability
		and mo.target and mo.target.valid and P_IsObjectOnGround(mo.target) and (mo.fuse == 1)
			TR.Game.spawnTronTrail(mo) -- This is spawned every tic, no need to P_RemoveMobj
		elseif (mo.type == MT_KAMIKAZESHIELD) -- pw_explodering ability
			if mo.target and mo.target.valid
			and mo.target.player and mo.target.player.valid
			
				local p = mo.target.player

				if (p.playerstate != PST_LIVE) -- Player not alive anymore?
				or not (p.pflags & PF_ATTACKDOWN) -- Or did the player let go of the button?
					-- Go poof!
					if (p.chargetimer > 3*TICRATE) -- Charged up enough to "Kamikaze" yourself?
						p.chargetimer = 0
						S_StopSound(mo.target)
						TR.Game.drainAmmo(p, pw_explosionring) -- Insult to injury
						S_StartSound(mo.target, sfx_s3k4e) -- 'Explosion'
												-- sfx_s3k54 is a good alternative

						-- Do AOE Explosion
						TR.Game.KamikazeExplode(p, 2*RING_DIST)
						
						if (p.playerstate != PST_DEAD) -- Double check to make sure the player isn't reeeealy dead.
							P_DamageMobj(mo.target, nil, nil, 1, DMG_DEATHMASK) -- Then kill the player.
						end
						P_StartQuake(FRACUNIT*6, 2)
						P_RemoveMobj(mo) -- Remove yourself.
						return -- Don't process anything else
					else -- Didn't charge up enough.
						p.chargetimer = 0
						S_StopSound(mo.target)
						S_StartSound(mo.target, sfx_kc65)
						P_RemoveMobj(mo) -- Remove yourself.
						return -- Don't process anything else
					end
				end

				if (p.chargetimer >= 4*TICRATE)
					p.chargetimer = 4*TICRATE
				else
					p.chargetimer = $ + 1
				end
				
				-- Set your position to your target.
				P_TeleportMove(mo, mo.target.x,
									mo.target.y,
									mo.target.z)

				--P_StartQuake(FRACUNIT*3, 2)

				TR.Game.WarnNearbyPlayersInRange(p, 2*RING_DIST)

				if (p.chargetimer >= 3*TICRATE)
					if (leveltime%8 == 0)
						S_StartSoundAtVolume(mo.target, sfx_s3k5c, 150)
					end
					
					if leveltime&1
						local gather = P_SpawnMobj(mo.x,mo.y,mo.z,MT_MSGATHER)
						gather.target = mo
						gather.fuse = 2
						gather.frame = $| FF_TRANS50
						gather.scale = FRACUNIT*2
					end
				elseif (p.chargetimer < 2*TICRATE)
					local gather = P_SpawnMobj(mo.x,mo.y,mo.z+mo.height/2,MT_KAMIKAZEGATHER)
					if gather and gather.valid then
						gather.target = mo
						gather.fuse = TICRATE
						gather.extravalue1 = P_RandomRange(0,359)*ANG1
						gather.extravalue2 = P_RandomRange(0,359)*ANG1
						gather.scale = mo.scale/4
					end
				end
			else
				P_RemoveMobj(mo)
				return
			end
		elseif (mo.type == MT_KAMIKAZEGATHER) -- pw_explodering gather 
			if mo.target and mo.target.valid
				local xyangle = mo.extravalue1
				local zangle = mo.extravalue2
				local dist = FRACUNIT*16*mo.fuse
				local xydist = P_ReturnThrustX(mo,zangle,dist)
				local zdist = P_ReturnThrustY(mo,zangle,dist)
				local x = mo.target.x + P_ReturnThrustX(mo,xyangle,xydist)
				local y = mo.target.y + P_ReturnThrustY(mo,xyangle,xydist)
				local z = mo.target.z + zdist + mo.target.height/2
				P_TeleportMove(mo,x,y,z)
			else
				P_RemoveMobj(mo)
				return
			end
		elseif (mo.type == MT_KAMIKAZEALERT)
			if mo.target and mo.target.valid
			and mo.target.player and mo.target.player.valid
				local p = mo.target.player
				if (p.playerstate == PST_LIVE)
				and p.kamialert
					-- Set your position to your target.
					P_TeleportMove(mo, mo.target.x,
										mo.target.y,
										mo.target.z)

					if (leveltime&1)
						mo.flags2 = $ | MF2_DONTDRAW
					else
						mo.flags2 = $ & ~(MF2_DONTDRAW)
					end
				else
					P_RemoveMobj(mo)
					return
				end
			end
		end
	end
end

TR.Game.fuseThinker = function(mo)
	if mo and mo.valid
		if (mo.type == MT_KAMIKAZEALERT)
			if mo.target and mo.target.valid
			and mo.target.player and mo.target.player.valid
				mo.target.player.kamialert = false
			end
		end
	end
end